/*
 * -------------------------------------------------------------
 *
 * television.c
 *
 *      is a program to display a live Image from the MV-1000
 *      television can have a own window or it  writes
 *      to a window which belongs to another program if
 *      it is called with a display and window_id
 *
 *      television.c supports 8Bit static color and
 *      8 Bit Pseudo color. It uses a private colortable
 *      but you will not even notice this normaly, because
 *      the private colortable is very similar to the  
 *      default colortable
 *
 *      television.c uses select() to wait for pictures, events
 *      from the X-Server, Events from the device driver and
 *      for input from the stdin pipe.
 *
 *
 * -------------------------------------------------------------
 */

#include <X11/Xlib.h>
#include <X11/X.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include "../mv1.h"
#include "../fglib.h"

#define NUM_COLS 64
#define NUM_ROI_COLS 5

XImage *xim;
Display *display;
Window window;
GC gc;
static int xsize = 0, ysize = 0;
static int xpos = 0, ypos = 0;
u_char *data;
XShmSegmentInfo shminfo;
Colormap cmap;
int screen_num;
Screen *screen;
Visual visual;
unsigned int depth;
u_long plane_masks[1];
int lut = -1;
int coltab[NUM_COLS];
XColor tofind[256];
XColor *roicol;
static MV1_GrabWindow gw;
static MV1_Frame frame;
MV1_Ctrl *bd;
static int shm = 0;
int zoom = 0;
int inverse = 0;
static int stopped = 0;
double gam = 1.5;
static int redraw_pending = 1;
int private_cmap = 1;
int coke_mode = 0;

void 
load_coke ()
{
	int i;
	for (i = 0; i < xsize * ysize; i++) {
		data[i] = i & 255;
	}
}

int
WinX (int x)
{
	return x >> zoom;
}
int 
WinDX (int dx)
{
	return dx >> zoom;
}
int 
WinY (int y)
{
	return y >> zoom;
}
int 
colorfunc (double i)
{
	double intens = pow ((i / (NUM_COLS - 1)), 1 / gam);
	if (inverse)
		intens = 1 - intens;
	return (int) 65535 *intens;
}

void 
recalc_colormap ()
{
	int i, j;
	u_long pixeltab[256];
	u_long luttable[256];
	Colormap defcmap = DefaultColormapOfScreen (screen);
	int num_cols = 0;
	for (j = 0; j < NUM_COLS; j++, num_cols++) {
		tofind[j].red = tofind[j].green = tofind[j].blue = colorfunc (j);
	}
	roicol = &tofind[num_cols];
	tofind[num_cols].red = 0;
	tofind[num_cols].green = 65535;
	tofind[num_cols].blue = 0;
	num_cols++;
	tofind[num_cols].red = 0;
	tofind[num_cols].green = 65535;
	tofind[num_cols].blue = 65535;
	num_cols++;
	tofind[num_cols].red = 65535;
	tofind[num_cols].green = 65535;
	tofind[num_cols].blue = 0;
	num_cols++;
	tofind[num_cols].red = 65535;
	tofind[num_cols].green = 0;
	tofind[num_cols].blue = 0;
	num_cols++;
	tofind[num_cols].red = 0;
	tofind[num_cols].green = 0;
	tofind[num_cols].blue = 32768;
	num_cols++;

	if (private_cmap) {
		FG_FindSimilarCmap (display, defcmap, cmap, tofind, num_cols, pixeltab);
	}
	else
		FG_TryGlobalCmap (display, tofind,
				  num_cols, pixeltab);

	for (j = 0; j < NUM_COLS; j++) {
		for (i = 0; i < 256 / NUM_COLS; i++) {
			luttable[i + (256 / NUM_COLS) * j] = pixeltab[j];
		}
	}
	// red for to bright areas and blue for to dark areas
	luttable[255] = roicol[NUM_ROI_COLS - 2].pixel;
	luttable[0] = roicol[4].pixel;
	if (depth == 8) {
		MV1_Write8BitLut (luttable, lut, bd);
	}
	else {
		MV1_WriteSoftLut8_32 (luttable, bd);
		printf ("hello\n");
	}

}


void 
grabproc ()
{
	//     memcpy(data,bd->video_mem ,xsize*ysize); 
	if ((!stopped) && (!coke_mode))
		if (MV1_XImageRead (&frame, xim, bd) <= 0)
			fprintf (stderr, "read error LINE %d\n", __LINE__);

	if (shm) {
		XShmPutImage (display, window, gc, xim, 0, 0, xpos, ypos, xsize, ysize, False);
	}
	else {
		XPutImage (display, window, gc, xim, 0, 0, xpos, ypos, xsize, ysize);
	}
	XSync (display, False);
	redraw_pending = 0;
}
/*
 * --------------------------------------------------------
 *
 * handle_input()
 *      read commands from a stdin pipe.
 *      This function is only called if stdin is not a tty
 *
 * --------------------------------------------------------
 */

void 
handle_input ()
{
	static char str[256];
	double r, x, y;
	while (fscanf (stdin, "%s %lf %lf %lf", str, &r, &x, &y) != EOF) {
		if ((strncmp (str, "start", 5) == 0)) {
			stopped = 0;
		}
		if ((strncmp (str, "stop", 4) == 0)) {
			stopped = 1;
		}
		else if (strncmp (str, "gamma", 5) == 0) {
			gam = r;
			if (gam < 0.01)
				gam = 0.01;
			inverse = 0;
			recalc_colormap ();
		}
		else if (strncmp (str, "invgamma", 8) == 0) {
			gam = r;
			if (gam < 0.01)
				gam = 0.01;
			inverse = 1;
			recalc_colormap ();
		}
		else if (strncmp (str, "expose", 6) == 0) {
			redraw_pending = 1;
		}
	}
}

void
find_visual (Visual * visual)
{
	XVisualInfo vTemplate;
	XVisualInfo *visualList;
	int visualsMatched;
	vTemplate.class = PseudoColor;
	vTemplate.screen = screen_num;
	visualList = XGetVisualInfo (display, VisualScreenMask | VisualClassMask,
				     &vTemplate, &visualsMatched);
	if (visualsMatched == 0) {
		private_cmap = 0;
		*visual = *DefaultVisual (display, screen_num);
		depth = DefaultDepth (display, screen_num);
	}
	else {
		*visual = *visualList[0].visual;
		depth = visualList[0].depth;
	}
	XFree (visualList);
}

int
main (int argc, char *argv[])
{
	int i;
	fd_set fds;
	u_long event_mask = 0;
	//  XEvent event;
	int Xfd;
	char *dpyname;
	struct timeval wait;
	bd = MV1_Open (0);
	if (bd == NULL)
		exit (1);
	MV1_SetMasterLoop(1,bd);
        MV1_SetMasterWaitFrame(1,bd); 
        MV1_SetTransferEndIrq(1,bd);
        MV1_SetMasterEnable(1,bd); // 

	MV1_InquireGrabWindow (&gw, bd);
//	fprintf(stderr,"Das gw ist dx %d, dy %d\n",gw.dx,gw.dy);
	MV1_CreateMaxFrame (&frame, &gw);
	MV1_SetFrame (&frame, bd);
//   MV1_InquireFrame(&frame,bd);
	xsize = frame.dx >> zoom;
	ysize = frame.dy >> zoom;

	if (argc < 3) {
		if ((display = XOpenDisplay (NULL)) == NULL) {
			fprintf (stderr, "can't open display \n");
			exit (1);
		}
		dpyname = XDisplayName (NULL);
	}
	else {
		if ((display = XOpenDisplay (argv[2])) == NULL) {
			fprintf (stderr, "can't open display %s\n", argv[2]);
			exit (1);
		}
		dpyname = argv[2];
	}
	if ((strcmp (dpyname, ":0.0") == 0) || (strcmp (dpyname, ":0.1") == 0)) {
		//fprintf(stderr,"shm !!\n");
		shm = 1;
	}

	screen_num = DefaultScreen (display);
	screen = DefaultScreenOfDisplay (display);
	find_visual (&visual);
	if (argc < 2) {
		window = XCreateSimpleWindow (display, RootWindow (display, screen_num), 0, 0,
					      xsize, ysize, 1,
					      BlackPixel (display, screen_num), WhitePixel (display, screen_num));
		gc = XCreateGC (display, window, 0, NULL);
	}
	else {
		Window dummywin;
		int parwidth, parheight;
		XRectangle clipRect;
		sscanf (argv[1], "%ld", &window);
		gc = XCreateGC (display, window, 0, NULL);
		XGetGeometry (display, window, &dummywin, &i, &i, &parwidth, &parheight, &i, &i);
		if (parwidth < xsize) {
			zoom = 1;
			MV1_SetZoom (bd, 1);
		}
		xsize = frame.dx >> zoom;
		ysize = frame.dy >> zoom;
		xpos = 0;	//(parwidth - xsize )/2;

		ypos = 0;	// (parheight - ysize )/2;

		clipRect.x = xpos;
		clipRect.y = xpos;
		clipRect.width = xsize;
		clipRect.height = ysize;
		XSetClipRectangles (display, gc, 0, 0, &clipRect, 1, Unsorted);
	}
	if (private_cmap) {
		u_long colors[256];
		event_mask |= ColormapChangeMask;
		cmap = XCopyColormapAndFree (display, DefaultColormapOfScreen (screen));
		XSetWindowColormap (display, window, cmap);
		XAllocColorCells (display, cmap, True, plane_masks, 0, colors, 256);
	}

	if (argc < 3) {
		event_mask |= ExposureMask;
	}
	XSelectInput (display, window, event_mask);
	Xfd = XConnectionNumber (display);
	if (depth == 8) {
		lut = MV1_GetPrivateLut (bd);
		if (lut < 0) {
			perror ("can't get private lut");
			lut = -1;
		}
	}
	if (shm) {
		xim = XShmCreateImage (display,
				       &visual,
			      depth, ZPixmap, NULL, &shminfo, xsize, ysize);
		shminfo.shmid = shmget (IPC_PRIVATE, xim->bytes_per_line * xim->height,
					IPC_CREAT | 0777);

		data = shminfo.shmaddr = xim->data = shmat (shminfo.shmid, 0, 0);
		shmctl (shminfo.shmid, IPC_RMID, NULL);
		shminfo.readOnly = False;
		XShmAttach (display, &shminfo);
	}
	else {
		data = malloc (xsize * ysize * depth / 8);
		xim = XCreateImage (display,
				    &visual, depth,
		      ZPixmap, 0, data, xsize, ysize, 8, xsize * depth / 8);
	}
	XMapWindow (display, window);
	recalc_colormap ();

	fcntl (STDIN_FILENO, F_SETFL, O_NONBLOCK);
	fcntl (STDOUT_FILENO, F_SETFL, O_NONBLOCK);
	while (1) {
		int recalc_colormap_pending = 0;
		i++;
	      repeat:
		if (!isatty (STDIN_FILENO)) {
			handle_input ();
		}
		while (XQLength (display)) {
			XEvent event;
			XNextEvent (display, &event);
			if (event.type == ColormapNotify) {
				recalc_colormap_pending = 1;
			}
		}
		if (recalc_colormap_pending) {
			recalc_colormap_pending = 0;
			recalc_colormap ();
		}
		grabproc ();
		if (XQLength (display) || redraw_pending)
			goto repeat;


		FD_ZERO (&fds);
		if (argc < 3) {
			FD_SET (Xfd, &fds);
		}
		if (!stopped)
			FD_SET (bd->raw_fd, &fds);
		if (!isatty (STDIN_FILENO))
			FD_SET (STDIN_FILENO, &fds);
		wait.tv_sec = 100;
		wait.tv_usec = 00000;
		select (32, &fds, NULL, NULL, /*&wait*/NULL);
/*
		fprintf(stderr,"r:%d c:%d s:%d \n",
			FD_ISSET(bd->raw_fd,&fds),
			FD_ISSET(bd->ctrl_fd,&fds),
			FD_ISSET(STDIN_FILENO,&fds),
		);
*/
	}
}
